 (***********************************************)
 (*                                             *)
 (*   Dinamic List v1.21 for Delphi 3,4,5       *)
 (*   1999-2000                                 *)
 (*   This component is freeware                *)
 (*                                             *)
 (*   Copiright by George Barbakadze            *)
 (*   All rights reserved                       *)
 (*                                             *)
 (***********************************************)

{             Model of accomodation of symbols in memory

 This model allows fast access to an elements of a list, for example:
 if in a list is loaded dictionary by 50000 word, for access with a
 properties of any word, program is required only pitches equally of
 length of this word (method GetFields).

                /\                                     /\
 |--------------|-------------TSymbolGroup-------------|--------------|
 |  |-----------|-------------|           |------------|-----------|  |
 |  | PrevGroup: PSymbolGroup |           |     PrevChar: char     |  |
 |  |-------------------------|           |------------------------|  |
 |                                                                    |
 |  |-------------------array[char] of TSymbol---------------------|  |
 |  |---------TSymbol---------|    |------|   |------|      |------|  |
 |  |            #0           |    |  #1  |   |  #2  |      | #255 |  |
 |  | Active: boolean         |    |      |   |      |      |      |  |
 |  | Structure: integer      |    |      |   |      | ...  |      |  |
 |  | Additional: pointer     |<-| |      |   |      |      |      |  |
 |  | ----------------------- |  | |      |   |      |      |      |  |
 |  | NextGroup: PSymbolGroup |  | |      |   |      |      |      |  |
 |  |-----|-------------------|  | |------|   |------|      |------|  |
 |--------|----------------------|------------------------------------|
          |     /\               |
          |     |                ----------------------|
         \/     |                                      |
 |--------------|-------------TSymbolGroup-------------|--------------|
 |  |-----------|-------------|           |------------|-----------|  |
 |  | PrevGroup: PSymbolGroup |           |     PrevChar: char     |  |
 |  |-------------------------|           |------------------------|  |
 |                                                                    |
 |  |-------------------array[char] of TSymbol---------------------|  |
 |  |---------TSymbol---------|    |------|   |------|      |------|  |
 |  |            #0           |    |  #1  |   |  #2  |      | #255 |  |
 |  | Active: boolean         |    |      |   |      |      |      |  |
 |  | Structure: integer      |    |      |   |      | ...  |      |  |
 |  | Additional: pointer     |    |      |   |      |      |      |  |
 |  | ----------------------- |    |      |   |      |      |      |  |
 |  | NextGroup: PSymbolGroup |    |      |   |      |      |      |  |
 |  |-----|-------------------|    |------|   |------|      |------|  |
 |--------|-----------------------------------------------------------|
          |
         \/

}

{$D+,L+,Y+}

unit DList;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  PSymbolGroup=^TSymbolGroup;
  TSymbol=record
    Active:boolean;
    Structure:integer;
    Additional:pointer;
    NextGroup:PSymbolGroup;
  end;
  TSymbolGroup=record
    PrevGroup:PSymbolGroup;
    PrevChar:char;
    Chars:array[char]of TSymbol;
  end;

  THeaderAction=(haRead,haWrite);

  TDisposeStructure=procedure(Sender: TObject;
    Structure: Integer; Properties: Pointer) of Object;
  TSaveStructure=procedure(Sender: TObject;
    Structure: Integer; Properties: Pointer; var F:file;
    Format: integer) of Object;
  TLoadStructure=procedure(Sender: TObject;
    Structure: Integer; var Properties: Pointer; var F:file;
    Format: integer) of Object;
  TFileHeader=procedure(Sender: TObject; var F:file;
    Action:THeaderAction; Format,InitPos: integer) of Object;

  TDList=class(Tcomponent)
  private
    { Private declarations }
    FSeparator:char;
    FOnDisposeStructure: TDisposeStructure;
    FOnSaveStructure: TSaveStructure;
    FOnLoadStructure: TLoadStructure;
    FOnFileHeader: TFileHeader;
    FirstGroup, ScrollGroup, BridgeGroup, TraceGroup: PSymbolGroup;
    ScrollChar, BridgeChar,TraceChar: char;
    ScrollLen,RootLen,Level1,Level2,CurLevel,IncLevel: integer;
    ScrollStr,Root: string;
    function GetSymbol(Str: string;var c:char):PSymbolGroup;
    function GetPart(Str:string;var c:char): PSymbolGroup;
    procedure ResetSymbolGroup(SymbolGroup: PSymbolGroup);
    procedure ClearSymbolGroup(var Group:PSymbolGroup;Level:integer);
    procedure ClearAllSymbolGroup(var Group:PSymbolGroup);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure SetFilter(RootStr:string;ForLevel,ToLevel:integer);
    procedure Add(Str:string;Struct:integer;Properties:pointer);
    function GetFields(Str:string;var Struct:integer;
      var Properties:pointer):boolean;
    procedure Remove(Str:string);
    procedure Clear;
    procedure ClearAll;
    function First:boolean;
    function Next:boolean;
    function GetName(var Str:string):boolean;
    function GetStructure(var Structure:integer):boolean;
    function GetProperties(var Properties:pointer):boolean;
    procedure BeginTrace(TraceRoot:string);
    function TraceNext(c:char;var Act:boolean;
                              var Properties:pointer):boolean;
    procedure Load(FileName: string; Format: integer);
    procedure Save(FileName: string; Format: integer);
  published
    { Published declarations }
    property Separator:char read FSeparator write FSeparator stored true;
    property OnDisposeStructure:TDisposeStructure
      read FOnDisposeStructure write FOnDisposeStructure;
    property OnSaveStructure:TSaveStructure
      read FOnSaveStructure write FOnSaveStructure;
    property OnLoadStructure:TLoadStructure
      read FOnLoadStructure write FOnLoadStructure;
    property OnFileHeader:TFileHeader
      read FOnFileHeader write FOnFileHeader;
  end;

implementation

constructor TDList.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FirstGroup:=nil;
  ScrollGroup:=nil;
  Separator := '_';
  SetFilter('',0,0);
end;

Destructor TDList.Destroy;
begin
  ClearAll;
  inherited destroy;
end;

procedure TDList.SetFilter(RootStr:string;ForLevel,ToLevel:integer);
begin
  Root:=RootStr;
  ScrollGroup:=nil;
  ScrollStr:='';
  RootLen:=Length(Root);
  Level1:=ForLevel;
  Level2:=ToLevel;
end;

procedure TDList.ResetSymbolGroup(SymbolGroup:PSymbolGroup);
var
  c:char;
begin
  for c:=Low(char) to High(char) do with SymbolGroup^.Chars[c] do begin
    Active:=false;
    Structure:=0;
    Additional:=nil;
    NextGroup:=nil;
  end;
end;

procedure TDList.Add(Str:string;Struct:integer;Properties:pointer);
var
  i:integer;
  c:char;
  CurGroup:PSymbolGroup;
begin
  if Str<>'' then begin
    CurGroup:=FirstGroup;
    if (CurGroup=nil)and(Length(Str)>0) then begin
      New(CurGroup);
      CurGroup^.PrevGroup:=nil;
      CurGroup^.PrevChar:=#0;
      FirstGroup:=CurGroup;
      ResetSymbolGroup(CurGroup);
    end;
    for i:=1 to Length(Str) do begin
      c:=Str[i];
      if i=Length(Str) then with CurGroup^.Chars[c] do begin
        try
          if Additional<>nil then if Assigned(FOnDisposeStructure) then
            FOnDisposeStructure(Self,Structure,Additional);
        finally
          Active:=true;
          Structure:=Struct;
          Additional:=Properties;
        end;
      end
      else begin
        if CurGroup^.Chars[c].NextGroup=nil then begin
          New(CurGroup^.Chars[c].NextGroup);
          ResetSymbolGroup(CurGroup^.Chars[c].NextGroup);
          CurGroup^.Chars[c].NextGroup^.PrevGroup:=CurGroup;
          CurGroup^.Chars[c].NextGroup^.PrevChar:=c;
        end;
        CurGroup:=CurGroup^.Chars[c].NextGroup;
      end;
    end;
  end
  else ShowMessage('Invalid name');
end;

function TDList.GetSymbol(Str:string;var c:char):PSymbolGroup;
var
  CurGroup:PSymbolGroup;
  i,Len:integer;
begin
  Result:=nil;
  Len:=Length(Str);
  CurGroup:=FirstGroup;
  c:=#0;
  i:=1;
  while (CurGroup<>nil)and(i<=Len) do with CurGroup^.Chars[Str[i]] do begin
    if Active and (i=Len) then begin
      Result:=CurGroup;
      c:=Str[i];
      Exit;
    end;
    CurGroup:=NextGroup;
    inc(i);
  end;
end;

function TDList.GetPart(Str:string;var c:char):PSymbolGroup;
var
  CurGroup:PSymbolGroup;
  i,Len:integer;
begin
  Result:=nil;
  Len:=Length(Str);
  CurGroup:=FirstGroup;
  c:=#0;
  i:=1;
  while (CurGroup<>nil)and(i<=Len) do with CurGroup^.Chars[Str[i]] do begin
    if i=Len then begin
      Result:=CurGroup;
      c:=Str[i];
      Exit;
    end;
    CurGroup:=NextGroup;
    inc(i);
  end;
end;

function TDList.GetFields(Str:string;var Struct:integer;
                                var Properties:pointer):boolean;
var
  Group:PSymbolGroup;
  c:char;
begin
  Result:=false;
  Group:=GetSymbol(Str,c);
  if Group<>nil then with Group^.Chars[c] do begin
    Result:=true;
    Struct:=Structure;
    Properties:=Additional;
  end;
end;

procedure TDList.Remove(Str:string);
var
  CurGroup,GroupSaver:PSymbolGroup;
  i:integer;
  CharFound,AllPassive:boolean;
  CharSaver,c:char;
begin
  if Str<>'' then begin
    CharSaver:=#0;
    GroupSaver:=nil;
    i:=1;
    CharFound:=true;
    CurGroup:=FirstGroup;
    while CharFound do begin
      if i>=Length(Str) then begin
        if CurGroup^.Chars[Str[i]].Active then begin
          CurGroup^.Chars[Str[i]].Active:=false;
          with CurGroup^.Chars[Str[i]] do if Additional<>nil then begin
            try
              if Assigned(FOnDisposeStructure) then
                FOnDisposeStructure(Self,Structure,Additional);
            finally
              Additional:=nil;
            end;
          end;
          CurGroup^.Chars[Str[i]].Structure:=0;
          CharFound:=false;
          AllPassive:=true;
          for c:=Low(char) to High(Char) do
            if CurGroup^.Chars[c].Active or
              (CurGroup^.Chars[c].NextGroup<>nil) then AllPassive:=false;
          if AllPassive then if CurGroup=FirstGroup then begin
            if FirstGroup=ScrollGroup then ScrollGroup:=nil;
            DisPose(FirstGroup);
            FirstGroup:=nil;
          end
          else begin
            if GroupSaver^.Chars[CharSaver].NextGroup=ScrollGroup then
              ScrollGroup:=nil;
            Dispose(GroupSaver^.Chars[CharSaver].NextGroup);
            GroupSaver^.Chars[CharSaver].NextGroup:=nil;
          end;
        end
        else CharFound:=false;
      end
      else begin
        GroupSaver:=CurGroup;
        CharSaver:=Str[i];
        CurGroup:=CurGroup^.Chars[Str[i]].NextGroup;
        if CurGroup=nil then CharFound:=false;
        inc(i);
      end;
    end;
  end;
end;

procedure TDList.ClearSymbolGroup(var Group:PSymbolGroup;Level:integer);
var
  c:char;
  AllPassive:boolean;
begin
  if Group<>nil then begin
    AllPassive:=true;
    for c:=Low(Char) to High(Char) do with Group^.Chars[c] do begin
      if c=Separator then IncLevel:=1 else IncLevel:=0;
      if (NextGroup<>nil)and(Level+IncLevel<=Level2) then
        ClearSymbolGroup(NextGroup,Level+IncLevel);
      if (Level+IncLevel>=Level1)and(Level+IncLevel<=Level2) then begin
        Active:=false;
        Structure:=0;
        if Additional<>nil then if Assigned(FOnDisposeStructure) then
          FOnDisposeStructure(Self,Structure,Additional);
        if NextGroup<>nil then AllPassive:=false;
      end
      else AllPassive:=false;
    end;
    if AllPassive then begin
      if ScrollGroup=Group then ScrollGroup:=nil;
      Dispose(Group);
      Group:=nil;
    end;
  end;
end;

procedure TDList.Clear;
var
  Group:PSymbolGroup;
  c:char;
begin
  if Root='' then ClearSymbolGroup(FirstGroup,0)
  else begin
    Group:=GetPart(Root,c);
    if Group<>nil then ClearSymbolGroup(Group^.Chars[c].NextGroup,0);
  end;
end;

procedure TDList.ClearAllSymbolGroup(var Group:PSymbolGroup);
var
  c:char;
begin
  if Group<>nil then begin
    for c:=Low(Char) to High(Char) do with Group^.Chars[c] do begin
      if NextGroup<>nil then ClearAllSymbolGroup(NextGroup);
      if Additional<>nil then begin
        try
          if Assigned(FOnDisposeStructure) then
            FOnDisposeStructure(Self,Structure,Additional);
        except
          Additional:=nil;
        end;
      end;
    end;
    if ScrollGroup=Group then ScrollGroup:=nil;
    Dispose(Group);
    Group:=nil;
  end;
end;

procedure TDList.ClearAll;
begin
  ClearAllSymbolGroup(FirstGroup);
end;

function TDList.First:boolean;
var
  Group:PSymbolGroup;
  c:char;
begin
  Result:=false;
  if Root='' then Group:=FirstGroup
  else begin
    Group:=GetPart(Root,c);
    if Group=nil then Exit
    else Group:=Group^.Chars[c].NextGroup;
  end;
  ScrollGroup:=Group;
  if Group=nil then Exit
  else begin
    ScrollChar:=#0;
    ScrollStr:=ScrollChar;
    ScrollLen:=RootLen;
    CurLevel:=0;
    IncLevel:=0;
    if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
    if ScrollGroup^.Chars[ScrollChar].Active and
      (CurLevel+IncLevel>=Level1)and
      (CurLevel+IncLevel<=Level2) then Result:=true
    else Result:=Next;
  end;
end;

function TDList.Next:boolean;
begin
  BridgeGroup:=ScrollGroup;
  BridgeChar:=ScrollChar;
  Result:=false;
  try
    if ScrollGroup<>nil then repeat
      while (ScrollGroup^.Chars[ScrollChar].NextGroup<>nil)and
        ((Level2>CurLevel)or(ScrollChar<>Separator)) do begin
        ScrollGroup:=ScrollGroup^.Chars[ScrollChar].NextGroup;
        if ScrollChar=Separator then inc(CurLevel);
        ScrollChar:=#0;
        ScrollStr:=ScrollStr+#0;
        inc(ScrollLen);
        if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
        if ScrollGroup^.Chars[ScrollChar].Active and
          (CurLevel+IncLevel>=Level1)and(CurLevel+IncLevel<=Level2) then begin
            Result:=true;
            Exit;
          end;
      end;
      while ScrollChar=High(char) do if RootLen=ScrollLen then Exit
      else begin
        ScrollChar:=ScrollGroup^.PrevChar;
        ScrollGroup:=ScrollGroup^.PrevGroup;
        ScrollStr:=Copy(ScrollStr,1,Length(ScrollStr)-1);
        dec(ScrollLen);
        if ScrollChar=Separator then dec(CurLevel);
      end;
      inc(ScrollChar);
      if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
      ScrollStr[Length(ScrollStr)]:=ScrollChar;
    until ScrollGroup^.Chars[ScrollChar].Active and
      (CurLevel+IncLevel>=Level1)and(CurLevel+IncLevel<=Level2)
    else Exit;
    Result:=true;
  finally
    if not Result then begin
      ScrollGroup:=BridgeGroup;
      ScrollChar:=BridgeChar;
    end;
  end;
end;

function TDList.GetName(var Str:string):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Str:=Root+ScrollStr;
    Result:=true;
  end;
end;

function TDList.GetStructure(var Structure:integer):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Structure:=ScrollGroup^.Chars[ScrollChar].Structure;
    Result:=true;
  end;
end;

function TDList.GetProperties(var Properties:pointer):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Properties:=ScrollGroup^.Chars[ScrollChar].Additional;
    Result:=true;
  end;
end;

procedure TDList.BeginTrace(TraceRoot:string);
begin
  if TraceRoot='' then TraceGroup:=nil
  else TraceGroup:=GetPart(TraceRoot,TraceChar);
end;

function TDList.TraceNext(c:char;var Act:boolean;
  var Properties:pointer):boolean;
begin
  if TraceGroup=nil then TraceGroup:=FirstGroup
  else TraceGroup:=TraceGroup^.Chars[TraceChar].NextGroup;
  if TraceGroup=nil then Result:=false
  else with TraceGroup^.Chars[c] do if Active or (NextGroup<>nil) then begin
    TraceChar:=c;
    Act:=Active;
    if Act then Properties:=Additional;
    Result:=true;
  end
  else Result:=false;
end;

procedure TDList.Save(FileName:string; Format:integer);
var
  f:file;
  BackupName,Str:string;
  Structure,i,Len:integer;
  Properties:pointer;
begin
{$i-}
  if IOResult=0 then begin end;
  if FileExists(FileName) then begin {Creation of backup file}
    BackupName:=ExtractFilePath(FileName)+'Backup of '+
      ExtractFileName(FileName);
    DeleteFile(BackupName);
    System.Assign(F,FileName);
    Rename(F,BackupName);
    if IOResult=0 then begin end;
  end;
  System.Assign(F,FileName);
  Rewrite(f,1);
  if IOResult=0 then begin
    try
      BlockWrite(F,Format,SizeOf(Format));
      if Assigned(FOnFileHeader) then
        FOnFileHeader(Self,F,haWrite,Format,SizeOf(Format));
      if First then repeat
        GetName(Str);
        Len:=Length(Str);
        BlockWrite(F,Len,SizeOf(Len));
        for i:=1 to Len do BlockWrite(F,Str[i],1);
        GetStructure(Structure);
        BlockWrite(F,Structure,SizeOf(Structure));
        GetProperties(Properties);
        if Assigned(FOnSaveStructure) then
          FOnSaveStructure(Self,Structure,Properties,F,Format);
      until not Next;
    finally
      System.Close(F);
    end;
  end;
end;

procedure TDList.Load(FileName:string; Format:integer);
var
  f:file;
  Str:string;
  ReadingFormat,Structure,i,Len:integer;
  Properties:pointer;
begin
{$i-}
  if IOResult=0 then begin end;
  System.Assign(F,FileName);
  Reset(F,1);
  if IOResult=0 then begin
    try
      BlockRead(F,ReadingFormat,SizeOf(Format));
      if ReadingFormat<=Format then begin
        if Assigned(FOnFileHeader) then
          FOnFileHeader(Self,F,haRead,ReadingFormat,SizeOf(ReadingFormat));
        while not Eof(F) do begin
          BlockRead(F,Len,SizeOf(Len));
          SetLength(Str,Len);
          for i:=1 to Len do BlockRead(F,Str[i],1);
          BlockRead(F,Structure,SizeOf(Structure));
          Properties:=nil;
          if Assigned(FOnLoadStructure) then
            FOnLoadStructure(Self,Structure,Properties,F,Format);
          Add(Str,Structure,Properties);
        end;
      end
      else ShowMessage('Uncnow format');
    finally
      System.Close(F);
    end;
  end;
end;

end.

